from __future__ import annotations

from typing import Protocol, TypeVar, Callable, Tuple, Any, Dict

from eventz.messages import Event, Command

T = TypeVar("T")


class ProcessesCommandsProtocol(Protocol):
    def process(self, command: Command) -> Tuple[Event, ...]:
        ...


class AppliesEventsProtocol(Protocol):
    def apply(self, event: Event) -> None:
        ...


class EmitsEventsProtocol(Protocol):
    def send(self, event: Event):
        ...


class RegistersEventHandlersProtocol(Protocol):
    def register_handler(self, name: str, handler: Callable):
        ...

    def deregister_handler(self, name: str):
        ...

    def list_handler_names(self) -> Tuple[str, ...]:
        ...

    def clear_handlers(self) -> None:
        ...


class EventBusProtocol(EmitsEventsProtocol, RegistersEventHandlersProtocol):
    pass


class MarshallProtocol(Protocol):
    def to_json(self, obj: Any) -> str:
        ...

    def from_json(self, json_string: str) -> Any:
        ...

    def register_codec(self, name: str, handler: MarshallCodecProtocol):
        ...

    def deregister_codec(self, name: str):
        ...


class MarshallCodecProtocol(Protocol):
    def serialise(self, obj: Any) -> Dict:
        ...

    def deserialise(self, params: Dict) -> Any:
        ...

    def handles(self, obj: Any) -> bool:
        ...


class JsonSerlialisable(Protocol):
    def get_json_data(self) -> Dict:
        ...



from typing import Callable, Tuple

from eventz.messages import Event
from eventz.protocols import EventBusProtocol


class EventBus:
    """
    Thin wrapper around the EventBusProtocol interface.
    Allows EventBus to be used statically/globally throughout the codebase
    whilst still keeping the actual implementation decoupled.
    """
    _implementation: EventBusProtocol = None

    @classmethod
    def set_implementation(cls, implementation: EventBusProtocol):
        cls._implementation = implementation

    @classmethod
    def send(cls, event: Event):
        return cls._implementation.send(event)

    @classmethod
    def register_handler(cls, name: str, handler: Callable):
        return cls._implementation.register_handler(name, handler)

    @classmethod
    def deregister_handler(cls, name: str):
        return cls._implementation.deregister_handler(name)

    @classmethod
    def list_handler_names(cls) -> Tuple[str, ...]:
        return cls._implementation.list_handler_names()

    @classmethod
    def clear_handlers(cls) -> None:
        print(cls._implementation)
        return cls._implementation.clear_handlers()



from typing import Optional

from eventz.entity import Entity


class Aggregate(Entity):
    """
    Currently a placeholder class where logic specific to root aggregate entities
    may be placed in the future.
    """
    def __init__(self, uuid: Optional[str] = None):
        super().__init__(uuid)

    @staticmethod
    def make_id() -> str:
        return Entity.make_id()



from typing import Sequence, Protocol, Tuple

from eventz.messages import Event


class EventStoreProtocol(Protocol):
    def fetch(self, aggregate_id: str) -> Tuple[Event, ...]:
        ...

    def persist(self, aggregate_id: str, events: Sequence[Event]):
        ...



from collections import Callable
from queue import SimpleQueue
from typing import Dict, Tuple

from eventz.messages import Event
from eventz.protocols import EventBusProtocol


class SubscriptionQueue:
    """
    FIFO subscription
    """
    def __init__(self):
        self._queue = SimpleQueue()

    def __call__(self, event: Event):
        self._queue.put(event)

    def get(self) -> Event:
        if not self._queue.empty():
            return self._queue.get()

    def all(self) -> Tuple[Event, ...]:
        events = []
        while not self._queue.empty():
            events.append(self._queue.get())
        return tuple(events)

    def __iter__(self):
        return self

    def __next__(self) -> Event:
        event = self.get()
        if event:
            return event
        raise StopIteration


class CollectAllHandler:
    def __init__(self, subscription: Callable):
        self._subscription: Callable = subscription

    def __call__(self, event: Event):
        self._subscription(event)


class OfTypeHandler:
    def __init__(self, subscription: Callable, of_type: type):
        self._subscription: Callable = subscription
        self._of_type: type = of_type

    def __call__(self, event: Event):
        if isinstance(event, self._of_type):
            self._subscription(event)


class EventBusDefault(EventBusProtocol):
    def __init__(self):
        self._handlers: Dict[str, Callable] = {}

    def send(self, event: Event):
        for handler in self._handlers.values():
            handler(event)

    def register_handler(self, name: str, handler: Callable):
        if name in self._handlers:
            err = f"Handler '{name}' already exists. Can't add it again."
            raise RuntimeError(err)
        self._handlers[name] = handler

    def deregister_handler(self, name: str):
        del self._handlers[name]

    def list_handler_names(self) -> Tuple[str, ...]:
        return tuple(self._handlers.keys())

    def clear_handlers(self) -> None:
        self._handlers = {}



from collections import Iterable

from eventz.protocols import ProcessesCommandsProtocol
from eventz.messages import Event, Command


class Service(ProcessesCommandsProtocol):
    def process(self, command: Command) -> Iterable[Event]:
        ...



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['eventz', 'eventz.event_store']

package_data = \
{'': ['*']}

install_requires = \
['immutables>=0.14,<0.15', 'python-dateutil>=2.8.1,<3.0.0']

setup_kwargs = {
    'name': 'eventz',
    'version': '0.1.0',
    'description': 'Eventz: an event-centric microframework.',
    'long_description': None,
    'author': 'Dan Ballance',
    'author_email': 'work@danballance.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
